package net.knarcraft.blacksmith.command.blacksmith;

import net.knarcraft.blacksmith.BlacksmithPlugin;
import net.knarcraft.blacksmith.command.ReloadCommand;
import net.knarcraft.blacksmith.config.SettingValueType;
import net.knarcraft.blacksmith.config.blacksmith.BlacksmithSetting;
import net.knarcraft.blacksmith.config.blacksmith.GlobalBlacksmithSettings;
import net.knarcraft.blacksmith.formatting.BlacksmithTranslatableMessage;
import net.knarcraft.blacksmith.formatting.ItemType;
import net.knarcraft.blacksmith.util.ConfigCommandHelper;
import net.knarcraft.blacksmith.util.InputParsingHelper;
import net.knarcraft.blacksmith.util.ItemHelper;
import net.knarcraft.blacksmith.util.TypeValidationHelper;
import org.bukkit.Material;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.enchantments.Enchantment;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Set;

/**
 * The command used for changing global blacksmith configuration options
 */
public class BlackSmithConfigCommand implements CommandExecutor {

    @Override
    public boolean onCommand(@NotNull CommandSender sender, @NotNull Command command, @NotNull String label,
                             @NotNull String[] args) {
        if (args.length == 0) {
            return false;
        }

        String commandName = args[0];
        if (commandName.equalsIgnoreCase("reload")) {
            return new ReloadCommand().onCommand(sender, command, label, args);
        }
        GlobalBlacksmithSettings settings = BlacksmithPlugin.getInstance().getGlobalBlacksmithSettings();

        //Find which setting the user has specified, if any
        BlacksmithSetting detectedSetting = BlacksmithSetting.getSetting(commandName);
        if (detectedSetting == null) {
            return false;
        }

        //Display the current value of a setting
        if (args.length == 1) {
            ConfigCommandHelper.displayCurrentValue(detectedSetting, settings, sender);
            return true;
        } else if (args.length == 2 && isSpecialCase(commandName)) {
            if (displaySpecialCaseValue(args[1], sender, detectedSetting, settings)) {
                return true;
            }
        }

        //Update the value of a special-case setting
        if (args.length == 3 && updateSpecialCase(settings, detectedSetting, args, sender)) {
            return true;
        }

        //Change the value of the specified setting
        if (TypeValidationHelper.isValid(detectedSetting.getValueType(), args[1], sender)) {
            ConfigCommandHelper.changeValue(args, detectedSetting, settings, sender);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Displays the current value of a special-case configuration value
     *
     * @param selector <p>The selector specifying which material/enchantment to display information about</p>
     * @param sender   <p>The sender to display the value to</p>
     * @param setting  <p>The setting to display</p>
     * @param settings <p>The settings object to query</p>
     * @return <p>True if the value was successfully displayed</p>
     */
    private boolean displaySpecialCaseValue(@NotNull String selector, @NotNull CommandSender sender,
                                            @Nullable BlacksmithSetting setting,
                                            @NotNull GlobalBlacksmithSettings settings) {
        if (setting == BlacksmithSetting.BASE_PRICE || setting == BlacksmithSetting.PRICE_PER_DURABILITY_POINT) {
            Material material = InputParsingHelper.matchMaterial(selector);
            if (material == null) {
                return false;
            }
            String currentValue;
            if (setting == BlacksmithSetting.BASE_PRICE) {
                currentValue = String.valueOf(settings.getBasePrice(material));
            } else {
                currentValue = String.valueOf(settings.getPricePerDurabilityPoint(material));
            }
            BlacksmithPlugin.getStringFormatter().displaySuccessMessage(sender,
                    BlacksmithTranslatableMessage.getItemCurrentValueMessage(setting.getCommandName(),
                            ItemType.MATERIAL, material.name(), currentValue));
            return true;
        } else if (setting == BlacksmithSetting.ENCHANTMENT_COST) {
            Enchantment enchantment = InputParsingHelper.matchEnchantment(selector);
            if (enchantment == null) {
                return false;
            }
            BlacksmithPlugin.getStringFormatter().displaySuccessMessage(sender,
                    BlacksmithTranslatableMessage.getItemCurrentValueMessage(setting.getCommandName(),
                            ItemType.ENCHANTMENT, enchantment.getKey().getKey(),
                            String.valueOf(settings.getEnchantmentCost(enchantment))));
            return true;
        } else {
            return false;
        }
    }

    /**
     * Gets whether a command name matches a special case command
     *
     * @param commandName <p>The command specified</p>
     * @return <p>True if the command is a special case</p>
     */
    private boolean isSpecialCase(@NotNull String commandName) {
        return commandName.equalsIgnoreCase(BlacksmithSetting.BASE_PRICE.getCommandName()) ||
                commandName.equalsIgnoreCase(BlacksmithSetting.PRICE_PER_DURABILITY_POINT.getCommandName()) ||
                commandName.equalsIgnoreCase(BlacksmithSetting.ENCHANTMENT_COST.getCommandName());
    }

    /**
     * Updates a special-case configuration value if a special-case is encountered
     *
     * @param settings          <p>The settings to modify</p>
     * @param blacksmithSetting <p>The setting specified</p>
     * @param args              <p>All arguments given</p>
     * @param sender            <p>The command sender to notify if successful</p>
     * @return <p>True if already handled as a special case</p>
     */
    private boolean updateSpecialCase(@NotNull GlobalBlacksmithSettings settings,
                                      @Nullable BlacksmithSetting blacksmithSetting,
                                      @NotNull String[] args,
                                      CommandSender sender) {
        if (InputParsingHelper.isEmpty(args[2])) {
            args[2] = "-1";
        } else if (!TypeValidationHelper.isValid(SettingValueType.POSITIVE_DOUBLE, args[2], sender)) {
            return true;
        }
        double newPrice = Double.parseDouble(args[2]);
        String newValue = String.valueOf(newPrice);

        if (blacksmithSetting == BlacksmithSetting.BASE_PRICE ||
                blacksmithSetting == BlacksmithSetting.PRICE_PER_DURABILITY_POINT) {
            return updatePriceSpecialCase(settings, blacksmithSetting, args[1], newPrice, sender);
        } else if (blacksmithSetting == BlacksmithSetting.ENCHANTMENT_COST) {
            //Update enchantment cost for an item
            Enchantment enchantment = InputParsingHelper.matchEnchantment(args[1]);
            if (enchantment == null) {
                return false;
            }
            ItemType itemType = ItemType.ENCHANTMENT;
            String itemChanged = enchantment.getKey().getKey();
            settings.setEnchantmentCost(enchantment, newPrice);
            BlacksmithPlugin.getStringFormatter().displaySuccessMessage(sender,
                    BlacksmithTranslatableMessage.getItemValueChangedMessage(blacksmithSetting.getCommandName(),
                            itemType, itemChanged, newValue));
            return true;
        } else {
            return false;
        }
    }

    /**
     * Updates a special case price configuration value if a special case is encountered
     *
     * @param settings          <p>The settings to modify</p>
     * @param blacksmithSetting <p>The setting specified</p>
     * @param materialName      <p>The material name to update the price for</p>
     * @param newPrice          <p>The new price to update to</p>
     * @param sender            <p>The command sender to respond to</p>
     * @return <p>True if the input was valid, and the item(s) was/were updated</p>
     */
    private boolean updatePriceSpecialCase(@NotNull GlobalBlacksmithSettings settings,
                                           @NotNull BlacksmithSetting blacksmithSetting,
                                           @NotNull String materialName, double newPrice,
                                           @NotNull CommandSender sender) {
        ItemType itemType = ItemType.MATERIAL;
        String itemChanged;
        //Update base price or price per durability point for an item
        if (materialName.contains("*")) {
            itemChanged = materialName;
            updateAllMatchedPrices(settings, blacksmithSetting, materialName, newPrice);
        } else {
            Material material = InputParsingHelper.matchMaterial(materialName);
            if (material == null) {
                return false;
            }
            itemChanged = material.name();
            if (blacksmithSetting == BlacksmithSetting.BASE_PRICE) {
                settings.setBasePrice(material, newPrice);
            } else {
                settings.setPricePerDurabilityPoint(material, newPrice);
            }
        }

        BlacksmithPlugin.getStringFormatter().displaySuccessMessage(sender,
                BlacksmithTranslatableMessage.getItemValueChangedMessage(blacksmithSetting.getCommandName(),
                        itemType, itemChanged, String.valueOf(newPrice)));
        return true;
    }

    /**
     * Updates all materials matching the material name wildcard
     *
     * @param settings          <p>The settings to modify</p>
     * @param blacksmithSetting <p>The setting specified</p>
     * @param materialName      <p>The wildcard material name to update cost for</p>
     * @param newPrice          <p>The new cost for the matched items</p>
     */
    private void updateAllMatchedPrices(@NotNull GlobalBlacksmithSettings settings,
                                        @NotNull BlacksmithSetting blacksmithSetting,
                                        @NotNull String materialName, double newPrice) {
        Set<Material> materials = ItemHelper.getWildcardMatch(materialName, false);
        for (Material material : materials) {
            if (blacksmithSetting == BlacksmithSetting.BASE_PRICE) {
                settings.setBasePrice(material, newPrice);
            } else {
                settings.setPricePerDurabilityPoint(material, newPrice);
            }
        }
    }

}
