package net.knarcraft.blacksmith.formatting;

import net.knarcraft.blacksmith.BlacksmithPlugin;

import java.util.Arrays;
import java.util.List;
import java.util.Random;

/**
 * A utility for formatting a string specifying an amount of time
 */
public final class TimeFormatter {

    private TimeFormatter() {

    }

    /**
     * Formats the given number of remaining seconds to the appropriate string
     *
     * @param exact   <p>Whether to display the exact number of seconds, or some unclear expression</p>
     * @param seconds <p>The number of seconds remaining</p>
     * @return <p>The time formatted correctly</p>
     */
    public static String formatTime(boolean exact, int seconds) {
        if (exact) {
            return net.knarcraft.knarlib.formatting.TimeFormatter.getDurationString(BlacksmithPlugin.getTranslator(), seconds);
        } else {
            return formatUnclearTime(seconds);
        }
    }

    /**
     * Gets a time format using vague wording to specify a known interval
     *
     * @param seconds <p>The number of seconds to format</p>
     * @return <p>Text describing approximate time remaining</p>
     */
    private static String formatUnclearTime(int seconds) {
        List<TimeInterval> intervals = Arrays.stream(TimeInterval.values()).sorted().toList();
        for (TimeInterval interval : intervals) {
            if (seconds < interval.getIntervalMax()) {
                return getMessageFromInterval(interval);
            }
        }
        return TimeInterval.INTERVAL_MORE_THAN_5_MINUTES.getDefaultText();
    }

    /**
     * Gets the text string corresponding to the given time interval
     *
     * @param interval <p>The time interval to get text for</p>
     * @return <p>Text describing the time interval</p>
     */
    private static String getMessageFromInterval(TimeInterval interval) {
        String text = BlacksmithPlugin.getStringFormatter().getUnFormattedMessage(
                BlacksmithTranslatableMessage.valueOf(interval.name()));

        //Choose a random entry if a comma-separated list is provided
        if (text.contains(",")) {
            String[] parts = text.split(",");
            String randomPart = parts[new Random().nextInt(parts.length)];
            if (randomPart != null) {
                text = randomPart;
            }
        }

        //Use the set message, or use the default
        if (!text.trim().isEmpty()) {
            return text;
        } else {
            return interval.getDefaultText();
        }
    }

}
