package net.knarcraft.blacksmith.util;

import net.knarcraft.blacksmith.BlacksmithPlugin;
import net.knarcraft.blacksmith.config.StargateYamlConfiguration;
import net.knarcraft.knarlib.property.ColorConversion;
import net.knarcraft.knarlib.util.FileHelper;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.configuration.MemorySection;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.configuration.file.YamlConfiguration;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * A helper class for getting an object value as the correct type
 */
public final class ConfigHelper {

    private ConfigHelper() {

    }

    /**
     * Gets the given value as a string list
     *
     * @param value <p>The raw string list value</p>
     * @return <p>The value as a string list, or null if not compatible</p>
     */
    public static @Nullable List<String> asStringList(@Nullable Object value) {
        if (value == null) {
            return new ArrayList<>();
        }
        if (value instanceof String string) {
            return List.of((string).split(","));
        } else if (value instanceof List<?> list) {
            List<String> strings = new ArrayList<>();
            for (Object object : list) {
                strings.add(String.valueOf(object));
            }
            return strings;
        } else {
            return null;
        }
    }

    /**
     * Gets the given value as a double
     *
     * <p>This will throw an exception if used for a non-double setting</p>
     *
     * @param value <p>The object value to get</p>
     * @return <p>The value of the given object as a double</p>
     * @throws ClassCastException    <p>If a non-double is given as input</p>
     * @throws NumberFormatException <p>If a non-double is given as input</p>
     */
    public static double asDouble(@NotNull Object value) throws ClassCastException, NumberFormatException {
        if (value instanceof String) {
            return Double.parseDouble((String) value);
        } else if (value instanceof Integer) {
            return (Integer) value;
        } else {
            return (Double) value;
        }
    }

    /**
     * Gets the given value as a boolean
     *
     * <p>This will throw an exception if used for a non-boolean value</p>
     *
     * @param value <p>The object value to get</p>
     * @return <p>The value of the given object as a boolean</p>
     * @throws ClassCastException <p>If the given value is not a boolean</p>
     */
    public static boolean asBoolean(@NotNull Object value) throws ClassCastException {
        if (value instanceof Boolean booleanValue) {
            return booleanValue;
        } else if (value instanceof String) {
            return Boolean.parseBoolean((String) value);
        } else {
            throw new ClassCastException();
        }
    }

    /**
     * Gets the given value as an integer
     *
     * <p>This will throw an exception if used for a non-integer setting</p>
     *
     * @param value <p>The object value to get</p>
     * @return <p>The value of the given object as an integer</p>
     * @throws ClassCastException <p>If the given value is not an integer</p>
     */
    public static int asInt(@NotNull Object value) throws ClassCastException {
        if (value instanceof String) {
            return Integer.parseInt((String) value);
        } else {
            return (Integer) value;
        }
    }

    /**
     * Changes all configuration values from the old name to the new name
     *
     * @param dataFolderPath       <p>The path to this plugin's data</p>
     * @param currentConfiguration <p>The current config to back up</p>
     */
    public static void migrateConfig(@NotNull String dataFolderPath, @NotNull FileConfiguration currentConfiguration) {
        BlacksmithPlugin instance = BlacksmithPlugin.getInstance();

        //Save the old config just in case something goes wrong
        try {
            currentConfiguration.save(new File(dataFolderPath, "config.yml.old"));
        } catch (IOException exception) {
            BlacksmithPlugin.warn("Unable to save old backup and do migration");
            return;
        }

        //Load old and new configuration
        instance.reloadConfig();
        FileConfiguration oldConfiguration = instance.getConfig();
        InputStream configStream = FileHelper.getInputStreamForInternalFile("/config.yml");
        if (configStream == null) {
            BlacksmithPlugin.error("Could not migrate the configuration, as the internal configuration could not be read!");
            return;
        }
        YamlConfiguration newConfiguration = StargateYamlConfiguration.loadConfiguration(
                FileHelper.getBufferedReaderFromInputStream(configStream));

        //Read all available config migrations
        Map<String, String> migrationFields;
        try {
            InputStream migrationStream = FileHelper.getInputStreamForInternalFile("/config-migrations.txt");
            if (migrationStream == null) {
                BlacksmithPlugin.error("Could not migrate the configuration, as the internal migration paths could not be read!");
                return;
            }
            migrationFields = FileHelper.readKeyValuePairs(FileHelper.getBufferedReaderFromInputStream(migrationStream),
                    "=", ColorConversion.NORMAL);
        } catch (IOException exception) {
            BlacksmithPlugin.warn("Unable to load config migration file");
            return;
        }

        //Replace old config names with the new ones
        for (String key : migrationFields.keySet()) {
            if (oldConfiguration.contains(key)) {
                migrateProperty(migrationFields, key, oldConfiguration);
            }
        }

        // Copy all keys to the new config
        for (String key : StargateYamlConfiguration.getKeysWithoutComments(oldConfiguration, true)) {
            if (oldConfiguration.get(key) instanceof MemorySection) {
                continue;
            }
            newConfiguration.set(key, oldConfiguration.get(key));
        }

        try {
            newConfiguration.save(new File(dataFolderPath, "config.yml"));
        } catch (IOException exception) {
            BlacksmithPlugin.warn("Unable to save migrated config");
        }

        instance.reloadConfig();
    }

    /**
     * Migrates one configuration property
     *
     * @param migrationFields  <p>The configuration fields to be migrated</p>
     * @param key              <p>The key/path of the property to migrate</p>
     * @param oldConfiguration <p>The original pre-migration configuration</p>
     */
    private static void migrateProperty(@NotNull Map<String, String> migrationFields, @NotNull String key,
                                        @NotNull FileConfiguration oldConfiguration) {
        String newPath = migrationFields.get(key);
        Object oldValue = oldConfiguration.get(key);
        if (!newPath.trim().isEmpty()) {
            if (oldConfiguration.isConfigurationSection(key)) {
                // Copy each value of a configuration section
                ConfigurationSection sourceSection = oldConfiguration.getConfigurationSection(key);
                ConfigurationSection destinationSection = oldConfiguration.createSection(newPath);
                if (sourceSection == null) {
                    return;
                }
                for (String path : StargateYamlConfiguration.getKeysWithoutComments(sourceSection, true)) {
                    destinationSection.set(path, sourceSection.get(path));
                }
            } else {
                // Copy the value to the new path
                oldConfiguration.set(newPath, oldValue);
            }
        }

        // Remove the old path's value
        oldConfiguration.set(key, null);
    }

}
